/**
 * @file arminstruction.h
 * @brief ARM 명령어 클래스 선언
 */

#ifndef ARMINSTRUCTION_H_
#define ARMINSTRUCTION_H_

#include<string>

using std::string;

struct ARMinstruction
{
public:
	void Decode(uint32_t binary);

	/// 명령어 문자열(mov, ldr, ...)
	string cmdString;
	/// 명령어 이진값을 저장
	uint32_t binary;
	/// condition : 명령어가 요구하는 상태(4비트), CPSR/SPSR의 첫 4비트와 같다.
	uint8_t cond;
	/// 27~25 세 비트를 딱히 지칭하는 단어가 없는듯 싶으므로.. 내 맘대로 붙이자.
	uint8_t codeId;
	/// 25비트를 따로 indirect bit로 쓰기도 한다.
	bool ibit;
	/// 24~21비트는 opcode에 해당한다. opcode가 아닌 다른 것일 가능성도 충분
	uint8_t opcode;
	/// 20비트는 S bit
	bool sbit;
	/// 19~16비트는 Rn 레지스터 번호
	uint8_t rn_num;
	/// 15~12비트는 Rd 레지스터 번호
	uint8_t rd_num;
	/// 11~7비트는 shift amount
	uint16_t shift_amount;
	/// 6,5비트는 shift
	uint8_t shift;
	// 4비트의 이름은 없다
	/// 3~0비트는 Rm 레지스터 번호
	uint8_t rm_num;
	/// 명령어 이름..
	string cmd;
	/// 명령어 이름(정수)
	uint8_t type;

private:
	void _decode0(uint32_t binary);
	void _decode1(uint32_t binary);
	void _decode2(uint32_t binary);
	void _decode3(uint32_t binary);

	void _decodeDataProcessingImmediateShift(uint32_t binary);
};

string cmdString(uint8_t opcode);

/*
 * ARM 명령어 목록
 */

enum InstructionEnum
{
	UNKNOWN, // 0
	/*
	 * 지원되는 명령어들
	 */
	ADC,
	ADD,
	AND,
	B,
	BL,
	BIC,
	BKPT,
	BLX1,	// ARMv5 이상, T 확장이 아닌 CPU에서 다르게 동작
	BLX2,	// ARMv5 이상, T 확장이 아닌 CPU에서 다르게 동작
	BX,		// ARMv5 이상, ARMv4 T 확장, T 확장이 아닌 CPU에서 다르게 동작
	CDP,
	CLZ,
	CMN,
	CMP,
	EOR,
	LDC,
	LDM1,
	LDM2,
	LDM3,
	LDR,
	LDRB,
	LDRBT,
	LDRD,	// ARMv5TE 이상, 단 ARMv5TExP는 제외
	LDRH,
	LDRSB,
	LDRSH,
	LDRT,
	MCR,
	MCRR,	// MCRR2: ARMv6 이상
	MLA,
	MOV,
	MRC,
	MRRC,	// MRRC2: ARMv6 이상
	MRS,
	MSR,
	MUL,
	MVN,
	ORR,
	PLD,
	QADD,
	QDADD,
	QDSUB,
	QSUB,
	RSB,
	RSC,
	SBC,
	SMLA_XY,
	SMLAL,
	SMLAL_XY,
	SMLAW_Y,
	SMUL_XY,
	SMULL,
	SMULW_Y,
	STC,
	STM1,
	STM2,
	STR,
	STRB,
	STRBT,
	STRD,	// 5TE, 5TExP X
	STRH,
	STRT,
	SUB,
	SWI,
	SWP,	// deprecated in 6
	SWPB,	// deprecated in 6
	TEQ,
	TST,
	UMLAL,
	UMULL,

	/*
	 * ARMv6에서 추가된 명령어, 즉 지원하지 않는 명령어
	 */
	BXJ,	// ARMv5TEJ 포함
	CPS,
	CPY,
	LDREX,
	PKHBT,
	PKHTB,
	QADD16,
	QADD8,
	QADDSUBX,
	QSUB16,
	QSUB8,
	QSUBADDX,
	REV,
	REV16,
	REVSH,
	RFE,
	SADD16,
	SADD8,
	SADDSUBX,
	SEL,
	SETEND,
	SHADD16,
	SHADD8,
	SHADDSUBX,
	SHSUB16,
	SHSUB8,
	SHSUBADDX,
	SMLAD,
	SMLALD,
	SMLSD,
	SMLSLD,
	SMMLA,
	SMMLS,
	SMMUL,
	SMUAD,
	SMUSD,
	SRS,
	SSAT,
	SSAT16,
	SSUB16,
	SSUB8,
	SSUBADDX,
	STREX,
	SXTAB,
	SXTAB16,
	SXTAH,
	SXTB,
	SXTB16,
	SXTH,
	UADD16,
	UADD8,
	UADDSUBX,
	UHADD16,
	UHADD8,
	UHADDSUBX,
	UHSUB16,
	UHSUB8,
	UHSUBADDX,
	UMAAL,
	UQADD16,
	UQADD8,
	UQADDSUBX,
	UQSUB16,
	UQSUB8,
	UQSUBADDX,
	USAD8,
	USADA8,
	USAT,
	USAT16,
	USUB16,
	USUB8,
	USUBADDX,
	UXTAB,
	UXTAB16,
	UXTAH,
	UXTB,
	UXTB16,
	UXTH,

	END_OF_ALL_INSTRUCTION
};

#endif /* ARMINSTRUCTION_H_ */
